/*@ Atomic exchange and compare-and-swap operations.
 *
 * Copyright (c) 2001 - 2023 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_ATOMIC_H
#define su_ATOMIC_H
#include <su/code.h>
su_USECASE_MX_DISABLED
#include <su/x-atomic.h> /* 1. */
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
/* atomic {{{ */
#if defined su_ATOMIC_IS_REAL || defined su_HAVE_MT
# define a_X(X,Y) EXPORT X CONCAT(su__atomic_g_xchg_,Y)(X ATOMIC *store, X nval);
a_X(u8,8) a_X(u16,16) a_X(u32,32) a_X(u64,64) a_X(uz,z) a_X(up,p)
# undef a_X
# define a_X(X,Y) EXPORT X CONCAT(su__atomic_g_busy_xchg_,Y)(X ATOMIC *store, X nval);
a_X(u8,8) a_X(u16,16) a_X(u32,32) a_X(u64,64) a_X(uz,z) a_X(up,p)
# undef a_X
# define a_X(X,Y) EXPORT boole CONCAT(su__atomic_g_cas_,Y)(X ATOMIC *store, X oval, X nval);
a_X(u8,8) a_X(u16,16) a_X(u32,32) a_X(u64,64) a_X(uz,z) a_X(up,p)
# undef a_X
# define a_X(X,Y) EXPORT void CONCAT(su__atomic_g_busy_cas_,Y)(X ATOMIC *store, X oval, X nval);
a_X(u8,8) a_X(u16,16) a_X(u32,32) a_X(u64,64) a_X(uz,z) a_X(up,p)
# undef a_X
#endif /* su_ATOMIC_IS_REAL || su_HAVE_MT */
#include <su/x-atomic.h> /* 2. */
#ifndef su_ATOMIC_IS_REAL
# define su_ATOMIC_IS_REAL 0
#endif
#ifndef su__ATOMIC_HAVE_FUNS /* {{{ */
# ifdef su_SOURCE_ATOMIC
#  error Combination will not work
# endif
# define a_X(X,Y) \
	INLINE X CONCAT(su__atomic_xchg_,Y)(X ATOMIC *store, X nval){return CONCAT(su__atomic_g_xchg_,Y)(store, nval);}
a_X(u8,8) a_X(u16,16) a_X(u32,32) a_X(u64,64) a_X(uz,z) a_X(up,p)
# undef a_X
# define a_X(X,Y) \
	INLINE X CONCAT(su__atomic_busy_xchg_,Y)(X ATOMIC *store, X nval)\
		{return CONCAT(su__atomic_g_busy_xchg_,Y)(store, nval);}
a_X(u8,8) a_X(u16,16) a_X(u32,32) a_X(u64,64) a_X(uz,z) a_X(up,p)
# undef a_X
# define a_X(X,Y) \
	INLINE boole CONCAT(su__atomic_cas_,Y)(X ATOMIC *store, X oval, X nval)\
		{return CONCAT(su__atomic_g_cas_,Y)(store, oval, nval);}
a_X(u8,8) a_X(u16,16) a_X(u32,32) a_X(u64,64) a_X(uz,z) a_X(up,p)
# undef a_X
# define a_X(X,Y) \
	INLINE void CONCAT(su__atomic_busy_cas_,Y)(X ATOMIC *store, X oval, X nval)\
		{CONCAT(su__atomic_g_busy_cas_,Y)(store, oval, nval);}
a_X(u8,8) a_X(u16,16) a_X(u32,32) a_X(u64,64) a_X(uz,z) a_X(up,p)
# undef a_X
#endif /* su__ATOMIC_HAVE_FUNS }}} */
#undef su__ATOMIC_HAVE_FUNS
INLINE u8 su_atomic_xchg_8(u8 ATOMIC *store, u8 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_xchg_8(store, nval);
}
INLINE u16 su_atomic_xchg_16(u16 ATOMIC *store, u16 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_xchg_16(store, nval);
}
INLINE u32 su_atomic_xchg_32(u32 ATOMIC *store, u32 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_xchg_32(store, nval);
}
INLINE u64 su_atomic_xchg_64(u64 ATOMIC *store, u64 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_xchg_64(store, nval);
}
INLINE uz su_atomic_xchg_z(uz ATOMIC *store, uz nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_xchg_z(store, nval);
}
INLINE up su_atomic_xchg_p(up ATOMIC *store, up nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_xchg_p(store, nval);
}
INLINE u8 su_atomic_busy_xchg_8(u8 ATOMIC *store, u8 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_xchg_8(store, nval);
}
INLINE u16 su_atomic_busy_xchg_16(u16 ATOMIC *store, u16 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_busy_xchg_16(store, nval);
}
INLINE u32 su_atomic_busy_xchg_32(u32 ATOMIC *store, u32 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_busy_xchg_32(store, nval);
}
INLINE u64 su_atomic_busy_xchg_64(u64 ATOMIC *store, u64 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_busy_xchg_64(store, nval);
}
INLINE uz su_atomic_busy_xchg_z(uz ATOMIC *store, uz nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_busy_xchg_z(store, nval);
}
INLINE up su_atomic_busy_xchg_p(up ATOMIC *store, up nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_busy_xchg_p(store, nval);
}
INLINE boole su_atomic_cas_8(u8 ATOMIC *store, u8 oval, u8 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_cas_8(store, oval, nval);
}
INLINE boole su_atomic_cas_16(u16 ATOMIC *store, u16 oval, u16 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_cas_16(store, oval, nval);
}
INLINE boole su_atomic_cas_32(u32 ATOMIC *store, u32 oval, u32 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_cas_32(store, oval, nval);
}
INLINE boole su_atomic_cas_64(u64 ATOMIC *store, u64 oval, u64 nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_cas_64(store, oval, nval);
}
INLINE boole su_atomic_cas_z(uz ATOMIC *store, uz oval, uz nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_cas_z(store, oval, nval);
}
INLINE boole su_atomic_cas_p(up ATOMIC *store, up oval, up nval){
	ASSERT_RET(store != NIL, FAL0);
	return su__atomic_cas_p(store, oval, nval);
}
INLINE void su_atomic_busy_cas_8(u8 ATOMIC *store, u8 oval, u8 nval){
	ASSERT_RET_VOID(store != NIL);
	su__atomic_busy_cas_8(store, oval, nval);
}
INLINE void su_atomic_busy_cas_16(u16 ATOMIC *store, u16 oval, u16 nval){
	ASSERT_RET_VOID(store != NIL);
	su__atomic_busy_cas_16(store, oval, nval);
}
INLINE void su_atomic_busy_cas_32(u32 ATOMIC *store, u32 oval, u32 nval){
	ASSERT_RET_VOID(store != NIL);
	su__atomic_busy_cas_32(store, oval, nval);
}
INLINE void su_atomic_busy_cas_64(u64 ATOMIC *store, u64 oval, u64 nval){
	ASSERT_RET_VOID(store != NIL);
	su__atomic_busy_cas_64(store, oval, nval);
}
INLINE void su_atomic_busy_cas_z(uz ATOMIC *store, uz oval, uz nval){
	ASSERT_RET_VOID(store != NIL);
	su__atomic_busy_cas_z(store, oval, nval);
}
INLINE void su_atomic_busy_cas_p(up ATOMIC *store, up oval, up nval){
	ASSERT_RET_VOID(store != NIL);
	su__atomic_busy_cas_p(store, oval, nval);
}
/* }}} */
#include <su/x-atomic.h> /* 3. */
C_DECL_END
#include <su/code-ou.h>
#endif /* su_ATOMIC_H */
/* s-itt-mode */
