/*@ Byte order swapping.
 *
 * Copyright (c) 2001 - 2023 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_BOSWAP_H
#define su_BOSWAP_H
#include <su/code.h>
#include <su/x-boswap.h> /* 1. */
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
/* boswap {{{ */
#include <su/x-boswap.h> /* 2. */
INLINE u16 su_boswap_16(u16 v) {return su__boswap_16(v);}
INLINE u32 su_boswap_32(u32 v) {return su__boswap_32(v);}
INLINE u64 su_boswap_64(u64 v) {return su__boswap_64(v);}
INLINE uz su_boswap_z(uz v) {return su_6432(su__boswap_64,su__boswap_32)(v);}
#undef a_X
#define a_X(X) if(su_BOM_IS_LITTLE()) v = su_CONCAT(su_boswap_,X)(v); return v
INLINE u16 su_boswap_big_16(u16 v) {a_X(16);}
INLINE u32 su_boswap_big_32(u32 v) {a_X(32);}
INLINE u64 su_boswap_big_64(u64 v) {a_X(64);}
INLINE uz su_boswap_big_z(uz v) {a_X(z);}
#undef a_X
#define a_X(X) if(su_BOM_IS_BIG()) v = su_CONCAT(su_boswap_,X)(v); return v
INLINE u16 su_boswap_little_16(u16 v) {a_X(16);}
INLINE u32 su_boswap_little_32(u32 v) {a_X(32);}
INLINE u64 su_boswap_little_64(u64 v) {a_X(64);}
INLINE uz su_boswap_little_z(uz v) {a_X(z);}
#undef a_X
INLINE u16 su_boswap_net_16(u16 v) {return su_boswap_big_16(v);}
INLINE u32 su_boswap_net_32(u32 v) {return su_boswap_big_32(v);}
INLINE u64 su_boswap_net_64(u64 v) {return su_boswap_big_64(v);}
INLINE uz su_boswap_net_z(uz v) {return su_boswap_big_z(v);}
/* }}} */
#include <su/x-boswap.h> /* 3. */
C_DECL_END
#include <su/code-ou.h>
#endif /* su_BOSWAP_H */
/* s-itt-mode */
