/*@ I/O and path flags (IOPF).
 *
 * Copyright (c) 2001 - 2023 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_IO_PATH_FLAGS_H
#define su_IO_PATH_FLAGS_H
#include <su/code.h>
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
/* flags {{{ */
enum su_iopf_access{
	su_IOPF_EXIST = 0,
	su_IOPF_EXEC = 1u<<0,
	su_IOPF_WRITE = 1u<<1,
	su_IOPF_READ = 1u<<4
};
enum{
	su_IOPF_ACCESS_MASK = su_IOPF_READ | su_IOPF_WRITE | su_IOPF_EXEC
};
enum su_iopf_at{
	su_IOPF_AT_NONE = 0,
	su_IOPF_AT_RMDIR = 1u<<0,
	su_IOPF_AT_EACCESS = 1u<<1,
	su_IOPF_AT_LNK_FOLLOW = 1u<<2,
	su_IOPF_AT_LNK_NOFOLLOW = 1u<<3
};
enum su_iopf_permission{
	su_IOPF_XOTH = 1u<<0,
	su_IOPF_WOTH = 1u<<1,
	su_IOPF_ROTH = 1u<<2,
	su_IOPF_RWXOTH = su_IOPF_XOTH | su_IOPF_WOTH | su_IOPF_ROTH,
	su_IOPF_XGRP = 1u<<3,
	su_IOPF_WGRP = 1u<<4,
	su_IOPF_RGRP = 1u<<5,
	su_IOPF_RWXGRP = su_IOPF_XGRP | su_IOPF_WGRP | su_IOPF_RGRP,
	su_IOPF_XUSR = 1u<<6,
	su_IOPF_WUSR = 1u<<7,
	su_IOPF_RUSR = 1u<<8,
	su_IOPF_RWXUSR = su_IOPF_XUSR | su_IOPF_WUSR | su_IOPF_RUSR
};
enum{
	su_IOPF_PERM_MASK = su_IOPF_RWXOTH | su_IOPF_RWXGRP | su_IOPF_RWXUSR,
	su_IOPF_PERM_DEF = su_IOPF_PERM_MASK & ~(su_IOPF_XOTH | su_IOPF_XGRP | su_IOPF_XUSR)
};
enum su_iopf_protection{
	su_IOPF_SVTX = 1u<<9,
	su_IOPF_SGID = 1u<<10,
	su_IOPF_SUID = 1u<<11
};
enum{
	su_IOPF_PROT_MASK = su_IOPF_SVTX | su_IOPF_SGID | su_IOPF_SUID
};
enum su_iopf_type{
	su_IOPF_FIFO = 1u<<12,
	su_IOPF_CHR = 2u<<12,
	su_IOPF_DIR = 4u<<12,
	su_IOPF_BLK = 6u<<12,
	su_IOPF_REG = 8u<<12,
	su_IOPF_LNK = 10u<<12,
	su_IOPF_SOCK = 12u<<12,
	su_IOPF_WHT = 14u<<12
};
enum{
	su_IOPF_TYPE_MASK = 15u<<12
};
/* }}} */
C_DECL_END
#include <su/code-ou.h>
#endif /* su_IO_PATH_FLAGS_H */
/* s-itt-mode */
