/*@ SipHash pseudorandom function (message authentication code).
 *
 * Copyright (c) 2021 - 2023 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC AND CC0-1.0
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_MD_SIPHASH_H
#define su_MD_SIPHASH_H
#include <su/code.h>
#if defined su_HAVE_MD || defined DOXYGEN /*XXX DOXYGEN bug; ifdef su_HAVE_MD*/
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
struct su_siphash;
/* su_siphash {{{ */
enum{
	su_SIPHASH_KEY_SIZE = 16u,
	su_SIPHASH_KEY_SIZE_MIN = su_SIPHASH_KEY_SIZE,
	su_SIPHASH_KEY_SIZE_MAX = su_SIPHASH_KEY_SIZE,
	su_SIPHASH_DIGEST_SIZE_64 = 8u,
	su_SIPHASH_DIGEST_SIZE_128 = 16u,
	su_SIPHASH_DIGEST_SIZE_MIN = su_SIPHASH_DIGEST_SIZE_64,
	su_SIPHASH_DIGEST_SIZE_MAX = su_SIPHASH_DIGEST_SIZE_128,
	su_SIPHASH_BLOCK_SIZE = 8u,
	su__SIPHASH_DEFAULT_CROUNDS = 2,
	su__SIPHASH_DEFAULT_DROUNDS = 4
};
enum su_siphash_digest{
	su_SIPHASH_DIGEST_64,
	su_SIPHASH_DIGEST_128
};
struct su_siphash{
	BITENUM_IS(u8,su_siphash_digest) sh_digest;
	u8 sh_compress_rounds;
	u8 sh_finalize_rounds;
	u8 sh__pad[1];
	u32 sh_carry_size;
	u64 sh_bytes;
	u64 sh_v0, sh_v1, sh_v2, sh_v3;
	u8 sh_carry[su_SIPHASH_BLOCK_SIZE];
};
EXPORT s32 su_siphash_setup_custom(struct su_siphash *self, void const *key, enum su_siphash_digest digest_type,
		u8 crounds, u8 drounds);
INLINE s32 su_siphash_setup(struct su_siphash *self, void const *key){
	ASSERT(self);
	return su_siphash_setup_custom(self, key, su_SIPHASH_DIGEST_64,
		su__SIPHASH_DEFAULT_CROUNDS, su__SIPHASH_DEFAULT_DROUNDS);
}
EXPORT void su_siphash_update(struct su_siphash *self, void const *dat, uz dat_len);
EXPORT void su_siphash_end(struct su_siphash *self, void *digest_store);
EXPORT void su_siphash_once(void *digest_store, enum su_siphash_digest digest_type, void const *key,
		void const *dat, uz dat_len, u8 crounds, u8 drounds);
/* }}} */
C_DECL_END
#include <su/code-ou.h>
#endif /* su_HAVE_MD */
#endif /* su_MD_SIPHASH_H */
/* s-itt-mode */
