/*@ Threads of execution.
 *
 * Copyright (c) 2001 - 2023 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_THREAD_H
#define su_THREAD_H
#include <su/code.h>
#include <su/x-thread.h> /* 1. */
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
struct su_thread;
/* thread {{{ */
enum su_thread_flags{
	su_THREAD_MAIN = 1u<<0
	/*su_THREAD_HAVE_CXX = 1u<<16 xxx for now just cast the C++ object! */
};
struct su_thread{
	struct{
		MT( u8 os[su__THREAD_IMPL_SIZE]; )
		u8 ATOMIC lck;
		u8 flags;
		u8 pad__[2];
		s32 err;
		char const *name;
		struct su_thread *wait_last; /* Lock-wait-suspension list */
		void *nydctl;
	}
#ifdef su__THREAD_IMPL_ALIGNMENT
		su_CC_ALIGNED(su__THREAD_IMPL_ALIGNMENT)
# undef su__THREAD_IMPL_ALIGNMENT
#endif
		t_;
};
#define su_THREAD_ERR_SCOPE_IN() \
do{\
	struct su_thread *su____thread_self__ = su_thread_self();\
	s32 su____thread_self__err__ = su____thread_self__->t_.err /**/
#define su_THREAD_ERR_SCOPE_OU() \
	su____thread_self__->t_.err = su____thread_self__err__;\
}while(0)
EXPORT_DATA struct su_thread su__thread_main;
#include <su/x-thread.h> /* 2. */
INLINE char const *su_thread_name(struct su_thread const *self){
	ASSERT(self);
	return self->t_.name;
}
INLINE s32 su_thread_err(struct su_thread const *self){
	ASSERT(self);
	return self->t_.err;
}
INLINE void su_thread_err_set(struct su_thread *self, s32 e){
	ASSERT(self);
	self->t_.err = e;
}
INLINE struct su_thread *su_thread_self(void) {return su__thread_self();}
INLINE s32 su_thread_get_err(void) {return su_thread_err(su_thread_self());}
INLINE void su_thread_set_err(s32 e) {su_thread_err_set(su_thread_self(), e);}
EXPORT void su_thread_yield(void);
#ifdef su_HAVE_MT
# error .
#endif /* su_HAVE_MT */
/* }}} */
#include <su/x-thread.h> /* 3. */
C_DECL_END
#include <su/code-ou.h>
#endif /* su_THREAD_H */
/* s-itt-mode */
